--[[
	Name:		TargetArmor
	Version:	1.4
	Author:		Blackmail (knightchris2001@yahoo.com)
		Copyright (c) 2006 Christopher A. Florin (Blackmail on Arthus)
	Description:	Displays estimated armor value for targeted players and mobs.
	URL:		http://www.curse-gaming.com/mod.php?addid=3210
]]

local MELEEPERCENT = "MeleePercent";
local MELEEHITS = "MeleeHits";
local RANGEDPERCENT = "RangedPercent";
local RANGEDHITS = "RangedHits";

local TARGETDAMAGE = "TargetDamage";
local TARGETPOINTS = "TargetPoints";

local configversion = 3;    --increment this any time the configuration format changes
local databaseversion = 3;  --increment this any time the database format changes

local AverageMeleeHit;
local AverageRangedHit;

local PlayerLevel = 1;
local PlayerName;

local TargetIsPlayer = true;
local TargetLevel = 70;
local TargetName = "Blackmail";

local TargetHealth = 0;
local TargetDamage = 0;
local TargetPPP = 0;

local lEventHandler = {};

local TargetArmor_DAMAGEDATA;

    if ( GetLocale() == "frFR" ) then
        TargetArmor_DAMAGEDATA =
            {
                { type = "Ranged", pattern = "Votre tire avec une arme \195\160 feu touche (.+) et inflige (%d+) points de d\195\169g\195\162ts.", target = 0, pts = 1 },
                { type = "Ranged", pattern = "Votre Tirer une fl\195\168che touche (.+) et inflige (%d+) points de d\195\169g\195\162ts.", target = 0, pts = 1 },
                { type = "Ranged", pattern = "Votre Tir \195\160 l\'arbal\195\168te touche (.+) et inflige (%d+) points de d\195\169g\195\162ts.", target = 0, pts = 1 },
    --            { type = "Ranged", pattern = "Your Auto Shot hits (.+) for (%d+).", target = 0, pts = 1 },
                { type = "Melee", pattern = "Vous touchez (.+) et infligez (%d+) points de d\195\169g\195\162ts.", target = 0, pts = 1 },

                { exit = true }
            };
    --elseif ( GetLocale()=="enUS" ) then
    else
        TargetArmor_DAMAGEDATA =
            {
                { type = "Ranged", pattern = "Your Shoot Bow hits (.+) for (%d+).", target = 0, pts = 1 },
                { type = "Ranged", pattern = "Your Shoot Crossbow hits (.+) for (%d+).", target = 0, pts = 1 },
                { type = "Ranged", pattern = "Your Shoot Gun hits (.+) for (%d+).", target = 0, pts = 1 },
                { type = "Ranged", pattern = "Your Auto Shot hits (.+) for (%d+).", target = 0, pts = 1 },
                { type = "Melee", pattern = "You hit (.+) for (%d+).", target = 0, pts = 1 },

                { exit = true }
            };
    end

--RegisterForSave("MobArmor");
--RegisterForSave("PlayerArmor");
--RegisterForSave("TargetArmorConfig");

--OLDDATATHRESHOLD:
--  When TotalHitpointChange is equal to or greater than this number, half of all the previously collected data is
--      automatically assumed old and thrown out.
--  The default of 1000 is the equivalent of 1000% damage reduced. (total of all hits)
local OLDDATATHRESHOLD = 1000;
--PERCENTAGEMODIFIER:
--  The higher this number, an accurate reading will come faster, but it will adjust over time slower.
--  The lower this number, the slower an accurate reading will come, but it will adjust over time faster.
local PERCENTAGEMODIFIER = 0.98;
--DAMAGEMODIFIER:
--  The higher this is, the slower the learning curve, but more accurate over time.
--  The lower this is, the faster the learning curve, but will take more time to be accurate.
--  **Basically, the lower this number is, the less old data counts toward a target's estimated hitpoints.
local DAMAGEMODIFIER = 0.98;
--DAMAGECONFIDENCE:
--  The higher this is, the more confidence is placed in large damage hits and lower confidence in small damage hits.
--  Ths lower this is, the less confidence is placed in large damage hits and higher confidence in small damage hits.
--  **Basically, for every health point change over 2, the DAMAGEMODIFIER is decreased by this amount.
local DAMAGECONFIDENCE = 0.005;

local function TargetArmor_Display()
    if ( not TargetName or TargetName == "" ) then HideUIPanel(TargetArmorFrame); return;
    else ShowUIPanel(TargetArmorFrame); end

    local DISPLAYTEXT;
    local lines;
    if ( TargetArmorConfig["showname"] == true ) then
        DISPLAYTEXT = "Level "..TargetLevel.." "..TargetName.." :\n"; lines = 1;
    else
        DISPLAYTEXT = ""; lines = 0;
    end

    local TotalRangedPercentage, TotalRangedHits;
    local TotalMeleePercentage, TotalMeleeHits;
    local TotalDamage, TotalPoints;

    if ( TargetIsPlayer == true ) then

        if ( not PlayerArmor[PlayerName][TargetName] ) then
            DISPLAYTEXT = DISPLAYTEXT.."Invalid player name.";
            TargetArmor_SetDisplay(DISPLAYTEXT, lines+1);
            return;
        end

        TotalRangedPercentage = PlayerArmor[PlayerName][TargetName][RANGEDPERCENT];
        TotalRangedHits = PlayerArmor[PlayerName][TargetName][RANGEDHITS];
        TotalMeleePercentage = PlayerArmor[PlayerName][TargetName][MELEEPERCENT];
        TotalMeleeHits = PlayerArmor[PlayerName][TargetName][MELEEHITS];
        TotalDamage = PlayerArmor[PlayerName][TargetName][TARGETDAMAGE];
        TotalPoints = PlayerArmor[PlayerName][TargetName][TARGETPOINTS];
    else
        local targetname = TargetName..":"..TargetLevel;

        if ( not MobArmor[PlayerName][targetname] ) then
            DISPLAYTEXT = DISPLAYTEXT.."Mob not found in database.\nPossible invalid name and/or level.";
            TargetArmor_SetDisplay(DISPLAYTEXT, lines+2);
            return;
        end

        TotalRangedPercentage = MobArmor[PlayerName][targetname][RANGEDPERCENT];
        TotalRangedHits = MobArmor[PlayerName][targetname][RANGEDHITS];
        TotalMeleePercentage = MobArmor[PlayerName][targetname][MELEEPERCENT];
        TotalMeleeHits = MobArmor[PlayerName][targetname][MELEEHITS];
        TotalDamage = MobArmor[PlayerName][targetname][TARGETDAMAGE];
        TotalPoints = MobArmor[PlayerName][targetname][TARGETPOINTS];
    end

    if ( TotalRangedHits < 1 and TotalMeleeHits < 1 ) then
        DISPLAYTEXT = DISPLAYTEXT.."NO ARMOR DATA.\n"; lines = lines + 1;
    else
    --FIRST STEP: CALCULATE ARMOR
        local pr = 0;
        if ( TotalMeleeHits > TotalRangedHits ) then
            pr = 1 - ( TotalMeleePercentage ) / ( TotalMeleeHits );
        else
            pr = 1 - ( TotalRangedPercentage ) / ( TotalRangedHits );
        end

        local ea = TargetArmor_PercentToArmor(pr, PlayerLevel);

        pr = floor(pr*10000+.5)/100;
        ea = floor(ea+.5);

        DISPLAYTEXT = DISPLAYTEXT.."EA: "..ea.." / "..pr.."%\n"; lines = lines + 1;
    end

--SECOND STEP: CALCULATE HITPOINTS IF WE'RE CONFIGURED TO
    if ( TargetArmorConfig["showhp"] == true ) then
        local text = "";

        if ( TotalPoints == 0 ) then
            DISPLAYTEXT = DISPLAYTEXT.."NO HITPOINT DATA.\n"; lines = lines + 1;
        else
            local damageperpoint = TotalDamage/TotalPoints;
            local estimatedhealth = floor(damageperpoint*(TargetHealth/UnitHealthMax("target")*100)+.5);
            local estimatedmaxhealth = floor(damageperpoint*100+.5);
            local maximumerror = floor(damageperpoint*.5+.5);

            DISPLAYTEXT = DISPLAYTEXT..estimatedhealth.."/"..estimatedmaxhealth.." Error: +/- "..maximumerror; lines = lines + 1;
        end
    end

    TargetArmor_SetDisplay(DISPLAYTEXT, lines);
end

local function TargetArmor_DisplayMouseover()
    if ( TargetArmorConfig["showmouseover"] == false ) then HideUIPanel(TargetArmorMouseoverFrame); return; end

    local MouseoverName = UnitName("mouseover");
    local MouseoverLevel = UnitLevel("mouseover");
    local MouseoverIsPlayer = UnitIsPlayer("mouseover");

    if ( not MouseoverName or MouseoverName == "" ) then HideUIPanel(TargetArmorMouseoverFrame); return;
    else ShowUIPanel(TargetArmorMouseoverFrame); end

    local DISPLAYTEXT;
    local lines;

    if ( TargetArmorConfig["showname"] == true ) then
        DISPLAYTEXT = "Level "..MouseoverLevel.." "..MouseoverName.." :\n"; lines = 1;
    else
        DISPLAYTEXT = ""; lines = 0;
    end

    local TotalRangedPercentage, TotalRangedHits;
    local TotalMeleePercentage, TotalMeleeHits;
    local TotalDamage, TotalPoints;

    if ( MouseoverIsPlayer == true ) then
        TotalRangedPercentage = PlayerArmor[PlayerName][MouseoverName][RANGEDPERCENT];
        TotalRangedHits = PlayerArmor[PlayerName][MouseoverName][RANGEDHITS];
        TotalMeleePercentage = PlayerArmor[PlayerName][MouseoverName][MELEEPERCENT];
        TotalMeleeHits = PlayerArmor[PlayerName][MouseoverName][MELEEHITS];
        TotalDamage = PlayerArmor[PlayerName][MouseoverName][TARGETDAMAGE];
        TotalPoints = PlayerArmor[PlayerName][MouseoverName][TARGETPOINTS];
    else
        local mouseovername = MouseoverName..":"..MouseoverLevel;

        TotalRangedPercentage = MobArmor[PlayerName][mouseovername][RANGEDPERCENT];
        TotalRangedHits = MobArmor[PlayerName][mouseovername][RANGEDHITS];
        TotalMeleePercentage = MobArmor[PlayerName][mouseovername][MELEEPERCENT];
        TotalMeleeHits = MobArmor[PlayerName][mouseovername][MELEEHITS];
        TotalDamage = MobArmor[PlayerName][mouseovername][TARGETDAMAGE];
        TotalPoints = MobArmor[PlayerName][mouseovername][TARGETPOINTS];
    end

    if ( TotalRangedHits < 1 and TotalMeleeHits < 1 ) then
        DISPLAYTEXT = DISPLAYTEXT.."NO ARMOR DATA.\n"; lines = lines + 1;
    else
    --FIRST STEP: CALCULATE ARMOR
        local pr = 0;
        if ( TotalMeleeHits > TotalRangedHits ) then
            pr = 1 - ( TotalMeleePercentage ) / ( TotalMeleeHits );
        else
            pr = 1 - ( TotalRangedPercentage ) / ( TotalRangedHits );
        end

        local ea = TargetArmor_PercentToArmor(pr, PlayerLevel);

        pr = floor(pr*10000+.5)/100;
        ea = floor(ea+.5);

        DISPLAYTEXT = DISPLAYTEXT.."EA: "..ea.." / "..pr.."%\n"; lines = lines + 1;
    end

--SECOND STEP: CALCULATE HITPOINTS IF WE'RE CONFIGURED TO
    if ( TargetArmorConfig["showhp"] == true ) then
        local text = "";

        if ( TotalPoints == 0 ) then
            DISPLAYTEXT = DISPLAYTEXT.."NO HITPOINT DATA.\n"; lines = lines + 1;
        else
            local damageperpoint = TotalDamage/TotalPoints;
            local estimatedhealth = floor(TargetHealth*damageperpoint+.5);
            local estimatedmaxhealth = floor(damageperpoint*UnitHealthMax("mouseover")+.5);
            local maximumerror = floor(damageperpoint*.5+.5);

            DISPLAYTEXT = DISPLAYTEXT..estimatedhealth.."/"..estimatedmaxhealth.." Error: +/- "..maximumerror; lines = lines + 1;
        end
    end

    TargetArmor_SetMouseoverDisplay(DISPLAYTEXT, lines);
end

function TargetArmor_SetDisplay(display, lines)
    TargetArmorFrame:SetHeight(15+(lines*10));
    TargetArmorText:SetHeight(15+(lines*10));
    TargetArmorText:SetText(display);
    if ( TargetArmorConfig["enabled"] == true ) then ShowUIPanel(TargetArmorFrame); end
end

function TargetArmor_SetMouseoverDisplay(display, lines)
    local x, y = GetCursorPosition();

    TargetArmorMouseoverFrame:SetHeight(15+(lines*10));
    TargetArmorMouseoverText:SetHeight(15+(lines*10));
    TargetArmorMouseoverText:SetText(display);
    if ( TargetArmorConfig["enabled"] == true and TargetArmorConfig["showmouseover"] == true ) then
        ShowUIPanel(TargetArmorMouseoverFrame);
    end
end

--------------------------------------------------------------------------------------------------
-- Utility functions
--  * Neither of these functions attempt to do any rounding for you.
--  * Percentages are a number between 0 and 1, 1 being 100%
--------------------------------------------------------------------------------------------------
function TargetArmor_ArmorToPercent(armor, playerlevel)
    local ar = armor / ( 85 * playerlevel + 400 );
    return ar / ( ar + 1 );
end

function TargetArmor_PercentToArmor(percent, playerlevel)
    return ( percent * ( 85 * playerlevel + 400 ) ) / ( 1 - percent );
end

function TargetArmor_ResetConfig()
    TargetArmorConfig = {};

    TargetArmorConfig["bgcolor"] = {};
    TargetArmorConfig["bgcolor"]["red"] = 0.15;
    TargetArmorConfig["bgcolor"]["green"] = 0.15;
    TargetArmorConfig["bgcolor"]["blue"] = 0.15;
    TargetArmorConfig["bgcolor"]["alpha"] = 1.0;

    TargetArmorConfig["bordercolor"] = {};
    TargetArmorConfig["bordercolor"]["red"] = 0.4;
    TargetArmorConfig["bordercolor"]["green"] = 0.4;
    TargetArmorConfig["bordercolor"]["blue"] = 0.4;
    TargetArmorConfig["bordercolor"]["alpha"] = 1.0;

    TargetArmorConfig["textcolor"] = {};
    TargetArmorConfig["textcolor"]["red"] = 1.0;
    TargetArmorConfig["textcolor"]["green"] = 1.0;
    TargetArmorConfig["textcolor"]["blue"] = 1.0;
    TargetArmorConfig["textcolor"]["alpha"] = 1.0;

    TargetArmorConfig["enabled"] = true;
    TargetArmorConfig["showname"] = true;

    --Config version 2 options:
    TargetArmorConfig["showhp"] = true;

    --Config version 3 options:
    TargetArmorConfig["showmouseover"] = true;

    TargetArmorConfig["version"] = configversion;

    TargetArmor_UpdateDisplayConfigs();
end

--------------------------------------------------------------------------------------------------
-- OnFoo functions
--------------------------------------------------------------------------------------------------
function TargetArmor_OnLoad()
	for key,value in lEventHandler do
		this:RegisterEvent(key);
	end

    SlashCmdList["TargetArmor"] = function(arg1) TargetArmor_SlashCommandHandler(arg1); end
    SLASH_TargetArmor1 = "/targetarmor";   --/ta is used by tracker assist; besides abbreviations are too easy to overlap

    TargetArmor_Display();
    TargetArmor_DisplayMouseover();
end

function TargetArmor_OnEvent(event)
	if (lEventHandler[event]) then
		lEventHandler[event](arg1, arg2, arg3, arg4);
	end
end

function TargetArmor_SlashCommandHandler(arg1)
    local s, e, command;
    local args = {};
	s, e, command, args[0], args[1], args[2], args[3], args[4], args[5] = string.find(arg1, "(.+) (.+) (.+) (.+) (.+)");
	if ( not args[5] ) then s, e, command, args[0], args[1], args[2], args[3], args[4]  = string.find(arg1, "(.+) (.+) (.+) (.+) (.+) (.+)"); end
	if ( not args[4] ) then s, e, command, args[0], args[1], args[2], args[3] = string.find(arg1, "(.+) (.+) (.+) (.+) (.+)"); end
	if ( not args[3] ) then s, e, command, args[0], args[1], args[2] = string.find(arg1, "(.+) (.+) (.+) (.+)"); end
	if ( not args[2] ) then s, e, command, args[0], args[1] = string.find(arg1, "(.+) (.+) (.+)"); end
	if ( not args[1] ) then s, e, command, args[0] = string.find(arg1, "(.+) (.+)"); end
	if ( not args[0] ) then command = arg1; end

    if ( command ) then command = string.lower(command); else command = string.lower(arg1); end

    if ( command == "bgcolor" ) then
        if ( not args[0] and not args[1] ) then return; end
        if ( args[0] == "red" ) then    --yes, I'm aware this if statement could be rewritten better but its clear what's supposed to happen this way.
            TargetArmorConfig["bgcolor"]["red"] = args[1]/255;
        elseif ( args[0] == "green" ) then
            TargetArmorConfig["bgcolor"]["green"] = args[1]/255;
        elseif ( args[0] == "blue" ) then
            TargetArmorConfig["bgcolor"]["blue"] = args[1]/255;
        elseif ( args[0] == "alpha" ) then
            TargetArmorConfig["bgcolor"]["alpha"] = args[1]/255;
        else
            if ( not args[2] ) then return; end
            if ( not args[3] ) then args[3] = 1.0; end
            TargetArmorConfig["bgcolor"]["red"] = args[0]/255;
            TargetArmorConfig["bgcolor"]["green"] = args[1]/255;
            TargetArmorConfig["bgcolor"]["blue"] = args[2]/255;
            TargetArmorConfig["bgcolor"]["alpha"] = args[3]/255;
        end
    elseif ( command == "bordercolor" ) then
        if ( not args[0] and not args[1] ) then return; end
        if ( args[0] == "red" ) then    --yes, I'm aware this if statement could be rewritten better but its clear what's supposed to happen this way.
            TargetArmorConfig["bordercolor"]["red"] = args[1]/255;
        elseif ( args[0] == "green" ) then
            TargetArmorConfig["bordercolor"]["green"] = args[1]/255;
        elseif ( args[0] == "blue" ) then
            TargetArmorConfig["bordercolor"]["blue"] = args[1]/255;
        elseif ( args[0] == "alpha" ) then
            TargetArmorConfig["bordercolor"]["alpha"] = args[1]/255;
        else
            if ( not args[0] and not args[1] and not args[2] ) then return; end
            if ( not args[3] ) then args[3] = 1.0; end
            TargetArmorConfig["bordercolor"]["red"] = args[0]/255;
            TargetArmorConfig["bordercolor"]["green"] = args[1]/255;
            TargetArmorConfig["bordercolor"]["blue"] = args[2]/255;
            TargetArmorConfig["bordercolor"]["alpha"] = args[3]/255;
        end
    elseif ( command == "disable" ) then
        HideUIPanel(TargetArmorFrame);
        TargetArmorConfig["enabled"] = false;
    elseif ( command == "displayplayer" or command == "showplayer" ) then
        if ( not args[0] ) then return; end
        local playername = string.upper(string.sub(args[0], 1, 1))..string.lower(string.sub(args[0], 2));
        TargetName = playername; TargetLevel = -1; TargetIsPlayer = true;
        TargetArmor_Display();
    elseif ( command == "displaymob" or command == "showmob" ) then
        if ( not args[0] and not args[1] ) then return; end
        if ( args[3] ) then
            TargetName = args[0].." "..args[1].." "..args[2];
            TargetLevel = tonumber(args[3]);
        elseif ( args[2] ) then
            TargetName = args[0].." "..args[1];
            TargetLevel = tonumber(args[2]);
        else
            TargetName = args[0];
            TargetLevel = tonumber(args[1]);
        end
        TargetIsPlayer = false;
        TargetArmor_Display();
    elseif ( command == "enable" ) then
        ShowUIPanel(TargetArmorFrame);
        TargetArmorConfig["enabled"] = true;
    elseif ( command == "hidename" ) then
	    TargetArmorConfig["showname"] = false;
	    TargetArmor_Display();
    elseif ( command == "resetconfig" ) then
    	TargetArmor_ResetConfig();
    elseif ( command == "showhp" ) then
        TargetArmorConfig["showhp"] = not TargetArmorConfig["showhp"];
	    TargetArmor_Display();
    elseif ( command == "showmouseover" ) then
    	TargetArmorConfig["showmouseover"] = not TargetArmorConfig["showmouseover"];
    	TargetArmor_DisplayMouseover();
    elseif ( command == "showname" ) then
	    TargetArmorConfig["showname"] = true;
	    TargetArmor_Display();
    elseif ( command == "textcolor" ) then
        if ( not args[0] and not args[1] ) then return; end
        if ( args[0] == "red" ) then    --yes, I'm aware this if statement could be rewritten better but its clear what's supposed to happen this way.
            TargetArmorConfig["textcolor"]["red"] = args[1]/255;
        elseif ( args[0] == "green" ) then
            TargetArmorConfig["textcolor"]["green"] = args[1]/255;
        elseif ( args[0] == "blue" ) then
            TargetArmorConfig["textcolor"]["blue"] = args[1]/255;
        elseif ( args[0] == "alpha" ) then
            TargetArmorConfig["textcolor"]["alpha"] = args[1]/255;
        else
            if ( not args[0] and not args[1] and not args[2] ) then return; end
            if ( not args[3] ) then args[3] = 1.0; end
            TargetArmorConfig["textcolor"]["red"] = args[0]/255;
            TargetArmorConfig["textcolor"]["green"] = args[1]/255;
            TargetArmorConfig["textcolor"]["blue"] = args[2]/255;
            TargetArmorConfig["textcolor"]["alpha"] = args[3]/255;
        end
    else
        DEFAULT_CHAT_FRAME:AddMessage("TargetArmor: Invalid command: "..command);
    end

    TargetArmor_UpdateDisplayConfigs();
end

function TargetArmor_UpdateDisplayConfigs()
    TargetArmorFrame:SetBackdropColor(TargetArmorConfig["bgcolor"]["red"], TargetArmorConfig["bgcolor"]["green"],
        TargetArmorConfig["bgcolor"]["blue"], TargetArmorConfig["bgcolor"]["alpha"]);
    TargetArmorMouseoverFrame:SetBackdropColor(TargetArmorConfig["bgcolor"]["red"], TargetArmorConfig["bgcolor"]["green"],
        TargetArmorConfig["bgcolor"]["blue"], TargetArmorConfig["bgcolor"]["alpha"]);

    TargetArmorFrame:SetBackdropBorderColor(TargetArmorConfig["bordercolor"]["red"], TargetArmorConfig["bordercolor"]["green"],
        TargetArmorConfig["bordercolor"]["blue"], TargetArmorConfig["bordercolor"]["alpha"]);
    TargetArmorMouseoverFrame:SetBackdropBorderColor(TargetArmorConfig["bordercolor"]["red"], TargetArmorConfig["bordercolor"]["green"],
        TargetArmorConfig["bordercolor"]["blue"], TargetArmorConfig["bordercolor"]["alpha"]);

    TargetArmorText:SetTextColor(TargetArmorConfig["textcolor"]["red"], TargetArmorConfig["textcolor"]["green"],
        TargetArmorConfig["textcolor"]["blue"], TargetArmorConfig["textcolor"]["alpha"]);
    TargetArmorMouseoverText:SetTextColor(TargetArmorConfig["textcolor"]["red"], TargetArmorConfig["textcolor"]["green"],
        TargetArmorConfig["textcolor"]["blue"], TargetArmorConfig["textcolor"]["alpha"]);

    if ( TargetArmorConfig["enabled"] == true ) then    --default layout in the xml code
        ShowUIPanel(TargetArmorFrame);
    else
        HideUIPanel(TargetArmorFrame);
    end

--    if ( TargetArmorConfig["showhp"] == true ) then
--    else
--    end

    if ( TargetArmorConfig["showmouseover"] == true ) then
        ShowUIPanel(TargetArmorMouseoverFrame);
    else
        HideUIPanel(TargetArmorMouseoverFrame);
    end

--    if ( TargetArmorConfig["showname"] == true ) then
--    else
--    end
end

lEventHandler["PLAYER_LOGIN"] = function ()
    PlayerName = UnitName("player");
    PlayerLevel = UnitLevel("player");

    if ( not MobArmor[PlayerName] ) then MobArmor[PlayerName] = {}; end
    if ( not PlayerArmor[PlayerName] ) then PlayerArmor[PlayerName] = {}; end

    TargetLevel = PlayerLevel;
    TargetName = PlayerName;
    TargetIsPlayer = true;
    TargetArmor_Display();

    TargetArmor_UpdateDisplayConfigs();
end

lEventHandler["VARIABLES_LOADED"] = function()
	if ( not MobArmor ) then
	    MobArmor = {};
	    MobArmor["version"] = databaseversion;
    end

    if ( MobArmor["version"] == 1 or MobArmor["version"] == nil ) then
        for index, value in MobArmor do
            if ( index ~= "version" ) then
                for index2, value2 in MobArmor[index] do
                    MobArmor[index][index2][MELEEPERCENT] = MobArmor[index][index2].MELEEPERCENT;
                    MobArmor[index][index2][MELEEHITS] = MobArmor[index][index2].MELEEHITS;
                    MobArmor[index][index2][RANGEDPERCENT] = MobArmor[index][index2].RANGEDPERCENT;
                    MobArmor[index][index2][RANGEDHITS] = MobArmor[index][index2].RANGEDHITS;

                    MobArmor[index][index2].MELEEPERCENT = nil;
                    MobArmor[index][index2].MELEEHITS = nil;
                    MobArmor[index][index2].RANGEDPERCENT = nil;
                    MobArmor[index][index2].RANGEDHITS = nil;
                end
            end
        end
        MobArmor["version"] = 2;
	end

    if ( MobArmor["version"] == 2 ) then
        for index, value in MobArmor do
            if ( index ~= "version" ) then
                for index2, value2 in MobArmor[index] do
                    MobArmor[index][index2][TARGETDAMAGE] = 0;
                    MobArmor[index][index2][TARGETPOINTS] = 0;
                end
            end
        end
        MobArmor["version"] = 3;
    end

    if ( not PlayerArmor ) then
        PlayerArmor = {};
        PlayerArmor["version"] = databaseversion;
    end

    if ( PlayerArmor["version"] == 1 ) then
        PlayerArmor["version"] = 2;
    end

    if ( PlayerArmor["version"] == 2 ) then
        for index, value in PlayerArmor do
            if ( index ~= "version" ) then
                for index2, value2 in PlayerArmor[index] do
                    PlayerArmor[index][index2][TARGETDAMAGE] = 0;
                    PlayerArmor[index][index2][TARGETPOINTS] = 0;
                end
            end
        end
        PlayerArmor["version"] = 3;
    end

	if ( not TargetArmorConfig ) then
	    TargetArmor_ResetConfig();
    elseif ( TargetArmorConfig["version"] == 1 ) then
        TargetArmorConfig["showhp"] = true;
    elseif ( TargetArmorConfig["version"] == 2 ) then
        TargetArmorConfig["showmouseover"] = true;
	end
	TargetArmorConfig["version"] = configversion;
end

lEventHandler["CHAT_MSG_COMBAT_SELF_HITS"] = function ()
    if ( not TargetName or TargetName == "" ) then return; end

    for index, value in TargetArmor_DAMAGEDATA do
        if ( value.exit ) then return; end

        local s, e;
        local results = { };
        s, e, results[0], results[1] = string.find(arg1, value.pattern);

        local match = false;
        if ( results[value.pts] and results[value.target] and results[value.target] == TargetName ) then match = true; end

        if ( match ) then
            if ( value.type == "Ranged" ) then
                --Do nothing; this is a melee event *SHOULD NOT HAPPEN*
            elseif ( value.type == "Melee" ) then
                local lowDmg, hiDmg, offlowDmg, offhiDmg, posBuff, negBuff, percentmod = UnitDamage("player");

                local currenthit = results[value.pts];  --Find the value of the last hit

                if ( ( offlowDmg ~= offhiDmg ) and ( offlowDmg ~= 0 or offhiDmg ~= 0 ) ) then
                    --If in here then the player is dualwielding.  We must not count the hits from offhand.
                    if ( tonumber(currenthit) < tonumber(offhiDmg) ) then return; end
                end

                AverageMeleeHit = (lowDmg + hiDmg) * percentmod / 2;    --Calculate the player's average melee hit
                local percentage = currenthit / AverageMeleeHit;

                local TotalMeleePercentage = 0;
                local TotalMeleeHits = 0;

                local OldDataModifier;

                --Retrieve old data
                if ( TargetIsPlayer == true ) then
                    TotalMeleePercentage = PlayerArmor[PlayerName][TargetName][MELEEPERCENT];
                    TotalMeleeHits = PlayerArmor[PlayerName][TargetName][MELEEHITS];
                    OldDataModifier = PERCENTAGEMODIFIER;
                else
                    local targetname = TargetName..":"..TargetLevel;

                    TotalMeleePercentage = MobArmor[PlayerName][targetname][MELEEPERCENT];
                    TotalMeleeHits = MobArmor[PlayerName][targetname][MELEEHITS];

                    if ( UnitLevel("player") == 60 ) then
                        OldDataModifier = 1;
                    else
                        OldDataModifier = PERCENTAGEMODIFIER;    --Mobs do not have evolving armor ratings, but players gain levels
                    end
                end

                if ( TotalMeleePercentage and TotalMeleeHits ) then   --If data exists, modify it
                    if ( TotalMeleePercentage > OLDDATATHRESHOLD ) then
                        TotalMeleePercentage = TotalMeleePercentage / 2;
                        TotalMeleeHits = TotalMeleeHits / 2;
                    end

                    TotalMeleePercentage = TotalMeleePercentage*OldDataModifier + percentage;
                    TotalMeleeHits = TotalMeleeHits*OldDataModifier + 1;
                else    --Set new data since no data exists
                    TotalMeleePercentage = percentage;
                    TotalMeleeHits = 1;
                end

                --Store new data
                if ( TargetIsPlayer == true ) then
                    PlayerArmor[PlayerName][TargetName][MELEEPERCENT] = TotalMeleePercentage;
                    PlayerArmor[PlayerName][TargetName][MELEEHITS] = TotalMeleeHits;
                else
                    local targetname = TargetName..":"..TargetLevel;

                    MobArmor[PlayerName][targetname][MELEEPERCENT] = TotalMeleePercentage;
                    MobArmor[PlayerName][targetname][MELEEHITS] = TotalMeleeHits;
                end

                --Calculate percentage decrease
                TargetArmor_Display();
                return;
            else
                --DEFAULT_CHAT_FRAME:AddMessage(value.type);
            end
        end
    end
end

lEventHandler["CHAT_MSG_SPELL_SELF_DAMAGE"] = function ()
    if ( not TargetName or TargetName == "" ) then return; end

    for index, value in TargetArmor_DAMAGEDATA do
        if ( value.exit ) then return; end

        local s, e;
        local results = { };
        s, e, results[0], results[1] = string.find(arg1, value.pattern);

        local match = false;
        if ( results[value.pts] and results[value.target] and results[value.target] == TargetName ) then match = true; end

        if ( match ) then
            if ( value.type == "Ranged" ) then
                local speed, lowDmg, hiDmg = UnitRangedDamage("player");    --Yes these two lines are ripped from WowWiki
                AverageRangedHit = (lowDmg + hiDmg) / 2;    --Calculate the player's average ranged hit

                local currenthit = results[value.pts];  --Find the value of the last hit
                local percentage = currenthit / AverageRangedHit;

                local TotalRangedPercentage = 0;
                local TotalRangedHits = 0;

                local OldDataModifier;

                --Retrieve old data
                if ( TargetIsPlayer == true ) then
                    TotalRangedPercentage = PlayerArmor[PlayerName][TargetName][RANGEDPERCENT];
                    TotalRangedHits = PlayerArmor[PlayerName][TargetName][RANGEDHITS];
                    OldDataModifier = PERCENTAGEMODIFIER;
                else
                    local targetname = TargetName..":"..TargetLevel;

                    TotalRangedPercentage = MobArmor[PlayerName][targetname][RANGEDPERCENT];
                    TotalRangedHits = MobArmor[PlayerName][targetname][RANGEDHITS];

                    if ( UnitLevel("player") == 60 ) then
                        OldDataModifier = 1;
                    else
                        OldDataModifier = PERCENTAGEMODIFIER;    --Mobs do not have evolving armor ratings, but players gain levels
                    end
                end

                if ( TotalRangedPercentage and TotalRangedHits ) then   --If data exists, modify it
                    if ( TotalRangedPercentage > OLDDATATHRESHOLD ) then
                        TotalRangedPercentage = TotalRangedPercentage / 2;
                        TotalRangedHits = TotalRangedHits / 2;
                    end

                    TotalRangedPercentage = TotalRangedPercentage*OldDataModifier + percentage;
                    TotalRangedHits = TotalRangedHits*OldDataModifier + 1;
                else    --Set new data since no data exists
                    TotalRangedPercentage = percentage;
                    TotalRangedHits = 1;
                end

                --Store new data
                if ( TargetIsPlayer == true ) then
                    PlayerArmor[PlayerName][TargetName][RANGEDPERCENT] = TotalRangedPercentage;
                    PlayerArmor[PlayerName][TargetName][RANGEDHITS] = TotalRangedHits;
                else
                    local targetname = TargetName..":"..TargetLevel;

                    MobArmor[PlayerName][targetname][RANGEDPERCENT] = TotalRangedPercentage;
                    MobArmor[PlayerName][targetname][RANGEDHITS] = TotalRangedHits;
                end

                --Calculate percentage decrease
                TargetArmor_Display();
                return;
            elseif ( value.type == "Melee" ) then
                --Do nothing, this is a Ranged event *SHOULD NOT HAPPEN*
            end
        end
    end
end

lEventHandler["PLAYER_TARGET_CHANGED"] = function ()
    TargetLevel = UnitLevel("target");
    TargetIsPlayer = UnitIsPlayer("target");
    TargetName = UnitName("target");
    TargetDamage = 0;
    TargetHealth = UnitHealth("target");
    TargetPPP = 0;

--    if ( not TargetIsPlayer ) then TargetIsPlayer = false; end
    if ( TargetName == nil or TargetName == "") then return; end

    if ( TargetIsPlayer == true ) then
        if ( not PlayerArmor[PlayerName][TargetName] ) then
            PlayerArmor[PlayerName][TargetName] = {};
            PlayerArmor[PlayerName][TargetName][RANGEDPERCENT] = 0;
            PlayerArmor[PlayerName][TargetName][RANGEDHITS] = 0;
            PlayerArmor[PlayerName][TargetName][MELEEPERCENT] = 0;
            PlayerArmor[PlayerName][TargetName][MELEEHITS] = 0;
            PlayerArmor[PlayerName][TargetName][TARGETDAMAGE] = 0;
            PlayerArmor[PlayerName][TargetName][TARGETPOINTS] = 0;
        end
    else
        local targetname = TargetName..":"..TargetLevel;

        if ( not MobArmor[PlayerName][targetname] ) then
            MobArmor[PlayerName][targetname] = {};
            MobArmor[PlayerName][targetname][MELEEPERCENT] = 0;
            MobArmor[PlayerName][targetname][MELEEHITS] = 0;
            MobArmor[PlayerName][targetname][RANGEDPERCENT] = 0;
            MobArmor[PlayerName][targetname][RANGEDHITS] = 0;
            MobArmor[PlayerName][targetname][TARGETDAMAGE] = 0;
            MobArmor[PlayerName][targetname][TARGETPOINTS] = 0;
        end
    end

     TargetArmor_Display();
end

lEventHandler["UPDATE_MOUSEOVER_UNIT"] = function ()
    if ( TargetArmorConfig["showmouseover"] == false ) then return; end

    local mouseovername = UnitName("mouseover");

    if ( mouseovername == nil or mouseovername == "" ) then HideUIPanel(TargetArmorMouseoverFrame); return; end

    if ( UnitIsPlayer("mouseover") == true ) then
        if ( PlayerArmor[PlayerName][mouseovername] ) then
            TargetArmor_DisplayMouseover();
            return;
        end
    else
        local mouseoverlevel = UnitLevel("mouseover");
        mouseovername = mouseovername..":"..mouseoverlevel;

        if ( MobArmor[PlayerName][mouseovername] ) then
            TargetArmor_DisplayMouseover();
            return;
        end
    end

    HideUIPanel(TargetArmorMouseoverFrame);
end

lEventHandler["UNIT_HEALTH"] = function ()
    if ( TargetName ~= UnitName("target") ) then return; end

	if( TargetName and arg1 == "target" ) then
        local health = UnitHealth("target");
        if ( UnitHealthMax("target") > 100 ) then    --Player is in party or something like that; we know its hp with 100% certainty
            if ( TargetIsPlayer == true ) then
                PlayerArmor[PlayerName][TargetName][TARGETDAMAGE] = UnitHealthMax("target");
                PlayerArmor[PlayerName][TargetName][TARGETPOINTS] = 100;
            else
                local targetname = TargetName..":"..TargetLevel;
                MobArmor[PlayerName][targetname][TARGETDAMAGE] = UnitHealthMax("target");
                MobArmor[PlayerName][targetname][TARGETPOINTS] = 100;
            end

            TargetDamage = 0;
            TargetHealth = health;

            TargetArmor_Display();

            return;
        elseif ( health and health > 0 and health < 100 and TargetDamage > 0 ) then
        -- I only track health less than 90% because you can hit something for 1 damage and it will show 99%, so
        -- this ensures that our data is at least "somewhat" accurate
            local change = abs(TargetHealth - health);
            local damage = TargetDamage;

            local totaldamage, totalhpchange;
            if ( TargetIsPlayer == true ) then
                totaldamage = PlayerArmor[PlayerName][TargetName][TARGETDAMAGE];
                totalhpchange = PlayerArmor[PlayerName][TargetName][TARGETPOINTS];

--[[
**********Begin Genetic damage algorithm
]]
                --First step: reduce data if it exceeds the threshold:
                if ( totalhpchange >= OLDDATATHRESHOLD ) then
                    totaldamage = totaldamage / 2;
                    totalhpchange = totalhpchange / 2;
                end

                --Second step: Adjust the amount of confidence we're placing in old data and a new hit recording based
                --  on the % of the target's hitpoints taken away.
                local tempmodifier = DAMAGEMODIFIER;
                local hitconfidence = DAMAGEMODIFIER + DAMAGECONFIDENCE*(change);
           	    if ( change == 0 ) then --no change in hitpoints, but damage was recorded, this is likely to skew our data.
                    damage = 0; tempmodifier = 1;
                elseif ( change == 1 ) then
           	        tempmodifier = 1 - (1 - tempmodifier) / 4;
                    hitconfidence = hitconfidence / 3;
                elseif ( change == 2 ) then
           	        tempmodifier = 1 - (1 - tempmodifier) / 3;
                    hitconfidence = hitconfidence / 2;
                elseif ( change == 3 ) then
                    tempmodifier = 1 - (1 - tempmodifier) / 2;
                    hitconfidence = hitconfidence / 1.2;
                elseif ( change >= 16 ) then
           	        tempmodifier = tempmodifier - 1.1*DAMAGECONFIDENCE*(change);
     	              hitconfidence = hitconfidence + 2*DAMAGECONFIDENCE*(change);
                elseif ( change >= 8 ) then
                    tempmodifier = tempmodifier - 1.05*DAMAGECONFIDENCE*(change);
                    hitconfidence = hitconfidence + DAMAGECONFIDENCE*(change);
                end

                --DEFAULT_CHAT_FRAME:AddMessage(tempmodifier.." : "..hitconfidence.." : "..damage.." : "..change);
                totaldamage = (totaldamage * tempmodifier) + (damage * hitconfidence);
                totalhpchange = (totalhpchange * tempmodifier) + (change * hitconfidence);
--[[
**********End Genetic damage algorithm
]]

            else
                local targetname = TargetName..":"..TargetLevel;
                totaldamage = MobArmor[PlayerName][targetname][TARGETDAMAGE] + damage;
                totalhpchange = MobArmor[PlayerName][targetname][TARGETPOINTS] + change;
            end

            if ( totalhpchange ~=0 and totaldamage ~= 0 ) then
                TargetPPP = totaldamage / totalhpchange;
            end

            if ( TargetIsPlayer == true ) then
                PlayerArmor[PlayerName][TargetName][TARGETDAMAGE] = totaldamage;
                PlayerArmor[PlayerName][TargetName][TARGETPOINTS] = totalhpchange;
            else
                local targetname = TargetName..":"..TargetLevel;
                MobArmor[PlayerName][targetname][TARGETDAMAGE] = totaldamage;
                MobArmor[PlayerName][targetname][TARGETPOINTS] = totalhpchange;
            end

            TargetDamage = 0;
            TargetHealth = health;
        else
            TargetDamage = 0;
            TargetHealth = UnitHealth("target");
        end

        TargetArmor_Display();
	end
end

lEventHandler["UNIT_COMBAT"] = function ()
    if ( arg4 == 0 ) then return; end
	if( TargetName and arg1 == "target" ) then
        if ( arg2 == "WOUND" ) then
            TargetDamage = TargetDamage + arg4;
        else
            TargetDamage = 0;
            TargetHealth = UnitHealth("target");
            return;
	    end

	end
end
